# encoding: UTF-8
#  dP"""8b    ii    88i    88   888888
#  Ib    "    __    88 i   88   88
#   '"~_      88    88  i  88   88888
#      ""b    88    88   i 88   88
#  "booodP    88    88    i88   888888

_VERSION = (0, 1, 1)

DESCRIPTION = 'the set of common tools'
URL = 'https://github.com/SineObama/'
VERSION = '.'.join(map(str, _VERSION))
REQUIRED = []
